from kafka import KafkaProducer

from demo.config import KAFKA_CONFIG


def produce_messages(n_messages):
	kafka_producer = KafkaProducer(
		bootstrap_servers=KAFKA_CONFIG['brokers'],
	)
	for i in range(n_messages):
		future = kafka_producer.send(KAFKA_CONFIG['topic'], b'raw_bytes')
		try:
			future.get(timeout=10)
		except Exception as e:
			print e
		kafka_producer.flush()


if __name__ == '__main__':
	produce_messages(1000)



import click

from logger import command_validator
from logger import grep_manager


@click.command()
@click.option('--topic', default=None, help='topic name')
@click.option('--brokers', default=None, help='broker cluster, separate with comma for each broker')
@click.option('--regex', default='', help='regex to be matched with the message')
def run_grep_command(topic, brokers, regex):
	try:
		command_validator.validate_grep_input(**locals())
		grep_manager.search_messages_in_parallel(topic, brokers, regex)
	except ValueError as e:
		print e




import uuid

from kafka import KafkaConsumer

kafka_consumer = None


def generate_kafka_consumer(brokers, is_singleton=True):
	"""
	:param str brokers: '129.0.0.1:2003,203.22.22.22:300'
	:param is_singleton: create only single reference when the flag is on
	:return:
	"""
	global kafka_consumer
	if not is_singleton:
		return _init_kafka_consumer(brokers)
	if kafka_consumer is None:
		kafka_consumer = _init_kafka_consumer(brokers)
	return kafka_consumer


def _init_kafka_consumer(brokers):
	group_id = 'debug-{}'.format(str(uuid.uuid1()))
	return KafkaConsumer(
		group_id=group_id,
		bootstrap_servers=[broker.strip(' ') for broker in brokers.split(',')],
		enable_auto_commit=False,
	)


def clear_kafka_consumer():
	"""
	Clear the singleton object
	:return:
	"""
	global kafka_consumer
	kafka_consumer = None



import click
from logger import grep_manager


@click.command()
@click.option('--topic', default='', help='topic name')
@click.option('--brokers', default='', help='localhost:9092,localhost:9093')
@click.option('--regex', default='', help='in case you are not familiar with regex, you can use free text like usual')
def run_grep_command(topic, brokers, regex):
	"""
	:param str topic:
	:param str brokers:
	:param str regex:
	:rtype: None
	"""
	grep_manager.search_messages_in_parallel(
		topic=topic,
		brokers=brokers,
		regex=regex,
	)


if __name__ == '__main__':
	run_grep_command()



import setuptools

long_description = ''

setuptools.setup(
	name='reverse-kafka-logger',
	version='0.1',
	packages=setuptools.find_packages(),
	long_description=open('README.md').read(),
	entry_points={
		'console_scripts': ['reverse-kafka-logger=logger.grep_interface:run_grep_command'],
	},
	install_requires=[
		'click',
	],
	author="Edward Sujono",
	description="pip install reverse-kafka-logger; kafka-logger --topic='gg' --brokers='localhost9092;localhost:2928' --regex='wow'",
	long_description_content_type="text/markdown",
	url="https://github.com/edwardsujono/kafka-logger",
	classifiers=[
		"Programming Language :: Python :: 2.7",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	],
)



import re
from collections import defaultdict
from multiprocessing import Process

from kafka import TopicPartition

from logger import kafka_factory
from logger.constant import BATCH_SIZE


def search_messages_in_parallel(topic, brokers, regex):
	"""
	Messages will be searched in parallel by spawning process per partition.
	:param topic:
	:param brokers:
	:param regex:
	:return:
	"""
	n_partition = _get_n_partition(brokers, topic)
	kafka_consumer = kafka_factory.generate_kafka_consumer(brokers)
	partition_id_to_start_end_offset = _get_partition_info(kafka_consumer, topic, n_partition)
	for partition in xrange(n_partition):
		p = Process(
			target=_reverse_search_log_per_partition,
			args=(brokers, topic, partition, partition_id_to_start_end_offset, regex),
		)
		p.start()
		p.join()


def _get_partition_info(kafka_consumer, topic, n_partition):
	partition_to_offset_info = defaultdict(dict)
	partitions = [TopicPartition(topic, partition) for partition in xrange(n_partition)]
	beginning_offsets = kafka_consumer.beginning_offsets(partitions)
	for topic_partition, offset in beginning_offsets.items():
		partition_to_offset_info[topic_partition.partition].update({'start_offset': offset})

	end_offsets = kafka_consumer.end_offsets(partitions)
	for topic_partition, offset in end_offsets.items():
		partition_to_offset_info[topic_partition.partition].update({'end_offset': offset})

	return partition_to_offset_info


def _reverse_search_log_per_partition(
	brokers,
	topic,
	partition,
	partition_id_to_start_end_offset,
	regex,
):
	"""
	This works by using a sliding window mechanism
	---------------------------
	1 2 3 4 5 6 7 8 9 10 11 12
							^
	Normal reading kafka starts from the beginning offset to the end
	we can seek the offset one by one, but there is an overhead of network
	to call the kafka broker, so the idea is to batch get the messages
	:param list[str] brokers:
	:param str topic:
	:param int partition:
	:param str regex:
	:return:
	"""
	"""
		Kafka consumer can only be instantiated when the sub-process is spawned otherwise the socket is closed
	"""
	kafka_consumer = kafka_factory.generate_kafka_consumer(brokers, is_singleton=False)
	start_offset = partition_id_to_start_end_offset[partition]['start_offset']
	end_offset = partition_id_to_start_end_offset[partition]['end_offset']
	print 'start_offset: {}, end_offset: {}'.format(start_offset, end_offset)
	kafka_consumer.assign([TopicPartition(topic, partition)])
	for offset in range(end_offset, start_offset - 1, -BATCH_SIZE):
		start_read_offset, end_read_offset = _get_start_end_offset(offset, start_offset)
		# assign partition and offset to the kafka consumer
		print 'start_read_offset: {}, end_read_offset: {}, assigned_offset: {}'.format(start_read_offset, end_read_offset, offset)
		kafka_consumer.seek(
			partition=TopicPartition(topic, partition),
			offset=start_read_offset,
		)
		grep_messages_in_batch(kafka_consumer, regex, start_read_offset, end_read_offset)


def _get_start_end_offset(offset, start_offset):
	"""
	start offset might be less than the offset that can be read. Depending with
	the configuration, messages are saved only in particular time period.
	:param offset:
	:param start_offset:
	:return:
	"""
	start_read_offset = offset - BATCH_SIZE
	end_read_offset = offset
	if start_read_offset < start_offset:
		start_read_offset = start_offset
	return start_read_offset, end_read_offset


def grep_messages_in_batch(kafka_consumer, regex, start_offset, end_offset):
	"""
	KafkaConsumer poll --> works by using intern
	:param KafkaConsumer kafka_consumer:
	:param str regex:
	:param int start_offset:
	:param int end_offset:
	:return:
	"""
	for _ in range(start_offset, end_offset):
		message = next(kafka_consumer)
		if re.match(regex, message.value):
			print 'message: {}'.format(message)


def _get_n_partition(brokers, topic):
	"""
	:param brokers:
	:param topic:
	:return:
	"""
	kafka_consumer = kafka_factory.generate_kafka_consumer(brokers, is_singleton=False)
	kafka_consumer.subscribe(topics=[topic])
	kafka_consumer.topics()
	return len(kafka_consumer.partitions_for_topic(unicode(topic)))
